from __future__ import annotations

import hashlib
import html
import importlib
import json
import logging
import os
import subprocess
import sys
import traceback
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any
from zoneinfo import ZoneInfo

import dill
import numpy as np
import pandas as pd
import pandas_market_calendars as mcal
import requests
from tabulate import tabulate

from minitrade.backtest import Backtest, Strategy
from minitrade.backtest.utils import calculate_positions
from minitrade.broker import Broker, BrokerAccount
from minitrade.datasource import QuoteSource
from minitrade.utils.config import config
from minitrade.utils.mailjet import mailjet_send_email
from minitrade.utils.mtdb import MTDB
from minitrade.utils.telegram import send_telegram_message

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

__all__ = [
    'TradePlan',
    'entry_strategy',
    'StrategyManager',
    'RawOrder',
    'BacktestLog',
    'BacktestRunner',
    'TraderLog',
    'Trader',
    'TaskManager',
    'TaskPlan',
    'TaskLog',
    'TaskRunner',
]


@dataclass(kw_only=True)
class TradePlan:
    '''TradePlan specifies how a strategy should be executed that includes:

    - which strategy to run 
    - the universe of assets as a list of tickers 
    - which data source to get price data from 
    - which timezone are the assets traded 
    - which date should a backtest starts 
    - which date should a trade order be generated from 
    - at what time of day should a backtest run
    - when order should be executed, i.e. on market open or on market close
    - which broker account should orders be submitted through 
    - how much initial cash should be invested
    '''

    __storage_root = os.path.expanduser('~/.minitrade/storage')

    id: str
    '''TradePlan ID'''

    name: str
    '''A unique human readable name'''

    strategy_file: str
    '''Strategy file name'''

    ticker_css: str
    '''Asset universe as a list of tickers in form of a comman separated string'''

    market_calendar: str
    '''Market calendar name as in pandas_market_calendars.get_calendar_names()'''

    market_timezone: str
    '''Timezone of the market where the tickers are traded'''

    data_source: str | None
    '''Quote source name as in QuoteSource.AVAILABLE_SOURCES'''

    backtest_start_date: str
    '''Since when backtest should start in string format "YYYY-MM-DD".'''

    trade_start_date: str | None
    '''Since when orders should be generated in string format "YYYY-MM-DD".
    Orders generated before the date are ignored.'''

    trade_time_of_day: str
    '''At what time of day should backtest starts in string format "HH:MM:SS".'''

    entry_type: str
    '''Type of order entry, 'TOO' for trade-on-open order, 'TOC' for trade-on-close order'''

    broker_account: str | None
    '''Alias of the broker account used to submit orders'''

    commission_rate: float = 0.001
    '''Commission ratio related to trade value'''

    initial_cash: float
    '''Cash amount to start with'''

    initial_holding: dict | None = None
    '''Asset positions to start with'''

    strict: bool = True
    '''If True, plan is traded in strict mode'''

    enabled: bool = False
    '''If the trade plan is enabled for trading'''

    create_time: datetime
    '''Time when the trade plan is created'''

    update_time: datetime | None
    '''Time when the trade plan is updated'''

    broker_ticker_map: dict | None = None
    '''A json string encoding a dict mapping from generic tickers to broker
    specific instrument IDs'''

    storage: dict | None = None
    '''A dict to store custom data from backtest runs'''

    def broker_instrument_id(self, ticker: str) -> Any:
        '''Get broker instrument ID corresponding to a generic `ticker`.

        Args:
            ticker: A generic ticker

        Returns:
            A broker specific instrument ID
        '''
        return self.broker_ticker_map[ticker]

    @staticmethod
    def list() -> list[TradePlan]:
        '''Return the list of available trade plans

        Returns:
            A list of zero or more trade plans
        '''
        plans = MTDB.get_all('TradePlan', orderby='name', cls=TradePlan)
        for plan in plans:
            plan.load_storage()
        return plans

    @staticmethod
    def get_plan(plan_id_or_name: str) -> TradePlan:
        '''Look up a trade plan by plan ID or plan name

        Args:
            plan_id_or_name: Trade plan ID or name

        Returns:
            Trade plan if found or None
        '''
        plan: TradePlan = (MTDB.get_one('TradePlan', 'id', plan_id_or_name, cls=TradePlan)
                           or MTDB.get_one('TradePlan', 'name', plan_id_or_name, cls=TradePlan))
        if plan:
            plan.load_storage()
        return plan

    def __call_scheduler(self, method: str, path: str, params: dict | None = None) -> Any:
        url = f'http://{config.scheduler.host}:{config.scheduler.port}{path}'
        resp = requests.request(method=method, url=url, params=params)
        if resp.status_code == 200:
            return resp.json()
        elif resp.status_code >= 400:
            raise RuntimeError(f'Scheduler {method} {url} {params} returns {resp.status_code} {resp.text}')

    def enable(self, enable: bool):
        '''Enable or disable a plan for trading.

        Args:
            enable: True to schedule a plan for trading, or False to deschedule.
        '''
        self.enabled = enable
        self.update_time = datetime.utcnow()
        MTDB.update('TradePlan', 'id', self.id, values={'enabled': self.enabled, 'update_time': self.update_time})
        self.__call_scheduler('PUT', f'/strategy/{self.id}')

    def jobinfo(self) -> dict:
        '''Get scheduled job status of the trade plan.

        Returns:
            A dict like {job_id: ..., job_frequency: ..., next_run_time: ...}
        '''
        return self.__call_scheduler('GET', f'/strategy/{self.id}')

    def save(self) -> None:
        '''Schedule the trade plan and save it to database.

        Fail if plan ID or plan name already exists.
        '''
        MTDB.save('TradePlan', self, blacklist=['storage'], on_conflict='error')
        self.__call_scheduler('PUT', f'/strategy/{self.id}')

    def delete(self) -> None:
        '''Unschedule a trade plan and delete it from database.
        '''
        MTDB.delete('TradePlan', 'id', self.id)
        try:
            os.remove(os.path.join(TradePlan.__storage_root, f'{self.id}.pkl'))
        except FileNotFoundError:
            pass
        self.__call_scheduler('DELETE', f'/strategy/{self.id}')

    def get_orders(self, run_id: str = None) -> list[RawOrder]:
        '''Retrieve all raw orders associated with a trade plan or only those generated during 
        a particular backtest run.

        Args:
            run_id: Backtest run ID. If None, return all orders generated from the trade plan.

        Returns:
            A list of zero or more RawOrders 
        '''
        if run_id:
            return MTDB.get_all('RawOrder', where={'plan_id': self.id, 'run_id': run_id}, cls=RawOrder)
        else:
            return MTDB.get_all('RawOrder', 'plan_id', self.id, cls=RawOrder)

    def cancel_pending_orders(self) -> None:
        '''Cancel all unsubmitted orders associated with a trade plan.
        '''
        orders = self.get_orders()
        for order in orders:
            if not order.broker_order_id:
                order.cancelled = True
                order.save()

    def list_logs(self, limit=30) -> list[BacktestLog]:
        '''Return all backtest history for this trade plan.

        Returns:
            A list of zero or more backtest logs
        '''
        return MTDB.get_all(
            'BacktestLog', 'plan_id', self.id, orderby=('log_time', False),
            limit=limit, cls=BacktestLog)

    def get_log(self, run_id: str) -> BacktestLog:
        '''Return backtest log for a particular run.

        Args:
            run_id: The backtest run ID

        Returns:
            Backtest log, or None if not found
        '''
        return MTDB.get_one('BacktestLog', 'id', run_id, cls=BacktestLog)

    def load_storage(self):
        '''Load storage from file.'''
        pklfile = os.path.join(TradePlan.__storage_root, f'{self.id}.pkl')
        if os.path.exists(pklfile):
            with open(pklfile, 'rb') as storage_file:
                self.storage = dill.load(storage_file)
        else:
            self.storage = {}

    def save_storage(self):
        '''Save storage to file.'''
        pklfile = os.path.join(TradePlan.__storage_root, f'{self.id}.pkl')
        with open(pklfile, 'wb') as storage_file:
            dill.dump(self.storage or {}, storage_file)


def entry_strategy(strategy):
    ''' Decorator to help specify the entry strategy when multiple strategy classes exists in a strategy file'''
    strategy.__entry_strategy__ = True
    return strategy


class StrategyManager:
    '''StrategyManager provides methods to organizes strategy files.'''

    __strategy_root = os.path.expanduser('~/.minitrade/strategy')

    @staticmethod
    def _path(filename: str) -> str:
        ''' Get the absolute path of a strategy file '''
        return os.path.join(StrategyManager.__strategy_root, filename)

    @staticmethod
    def list() -> list[str]:
        '''Return the list of available strategy files

        Returns:
            A list of strategy filenames
        '''
        return [f for f in os.listdir(StrategyManager.__strategy_root) if os.path.isfile(StrategyManager._path(f))]

    @staticmethod
    def save(filename: str, content: str):
        '''Save `content` to a strategy file. 

        Args:
            filename: Strategy file name
            content: File content as string
        '''
        module_file = StrategyManager._path(filename)
        with open(module_file, 'w') as f:
            f.write(content)

    @staticmethod
    def load(strategy_file: str) -> Strategy:
        '''Load strategy class from `strategy_file`.

        Args:
            strategy_file: The strategy file name

        Returns:
            The loaded strategy class

        Raises:
            RuntimeError: If a strategy class is not found for any reason.
        '''
        strategy_path = StrategyManager._path(strategy_file)
        module_name = strategy_file.removesuffix('.py')
        # load the strategy file
        spec = importlib.util.spec_from_file_location('strategy', strategy_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        # look up and return the strategy class
        strategy_found = []
        for attr in dir(module):
            attr_val = getattr(module, attr)
            try:
                if issubclass(attr_val, Strategy) and attr != 'Strategy':
                    strategy_found.append(attr_val)
            except Exception:
                pass
        if len(strategy_found) > 1:
            entry = [s for s in strategy_found if '__entry_strategy__' in s.__dict__]
            if len(entry) == 1:
                return entry[0]
            raise RuntimeError(
                f'Multiple strategy classes found: {strategy_found}. Use @entry_strategy decorator to specify which one should run.')
        elif len(strategy_found) == 1:
            return strategy_found[0]
        else:
            raise RuntimeError(f'No strategy found in {strategy_path}')

    @staticmethod
    def read(strategy_file: str) -> str:
        ''' Return the content of `strategy_file` 

        Args:
            strategy_file: The strategy file name

        Returns:
            The file content as a string
        '''
        strategy_path = StrategyManager._path(strategy_file)
        with open(strategy_path, 'r') as f:
            return f.read()

    @staticmethod
    def delete(strategy_file: str) -> None:
        '''Delete `strategy_file`

        Args:
            strategy_file: The strategy file name
        '''
        strategy_path = StrategyManager._path(strategy_file)
        os.remove(strategy_path)


@dataclass(kw_only=True)
class RawOrder:
    '''RawOrder represents an order generated during the backtest run of a strategy.
    It captures the asset, side, size, and timing of an order among others.
    The order is broker independent until it's submitted via a particular broker
    account, after which a broker order ID can be associated.
    '''

    id: str
    '''Raw order ID'''

    plan_id: str
    '''ID of the trade plan that generates this order'''

    run_id: str
    '''ID of particular backtest run that generates this order'''

    ticker: str
    '''Ticker to buy or sell'''

    side: str
    '''Side to take for the order, "Buy" or "Sell".'''

    size: int
    '''Number of shares to buy or sell, positive for long and negative for short.'''

    signal_time: datetime
    '''Date on which the order signal is generated.'''

    cancelled: bool = False
    '''True if the order is cancelled before submitting.'''

    broker_order_id: str | None
    '''Broker assigned order ID after the raw order is submitted.'''

    def save(self):
        '''Save self to database. 

        Overwrite existing ones if order id already exists.
        '''
        MTDB.save('RawOrder', self, on_conflict='update')

    @property
    def tag(self):
        '''A short descriptive label'''
        return f'{self.side} {self.ticker} {abs(self.size)}'


@dataclass(kw_only=True)
class BacktestLog:
    '''BacktestLog captures the inputs and outputs of a backtest run for diagnosis.'''

    id: str
    '''Run ID'''

    plan_id: str
    '''Trade plan ID'''

    plan_name: str
    '''Trade plan name'''

    plan_strategy: str
    '''Strategy filename'''

    plan: str
    '''Trade plan seralized in Json'''

    data: pd.DataFrame | None = None
    '''Quote data used in backtest'''

    strategy_code: str | None
    '''Strategy code snapshot'''

    params: str | None
    '''Backtest parameters'''

    result: pd.Series | None = None
    '''Backtest result'''

    exception: str | None = None
    '''Exception happened during backtesting'''

    stdout: str | None = None
    '''STDOUT capture during backtesting'''

    stderr: str | None = None
    '''STDERR capture during backtesting'''

    log_time: datetime
    '''Log time'''

    @property
    def error(self) -> bool:
        '''A backtest is regarded as failed if any of the following occurs:

        - Backtest generates no result.
        - Exception occurs.
        - STDERR output is not empty.

        Returns:
            True for error, False for success
        '''
        return bool(self.result is None or self.exception or self.stderr)

    def save(self):
        MTDB.save('BacktestLog', self, on_conflict='error')


class BacktestRunner:
    '''BacktestRunner handles the execution of a backtest, which involves preparing data,
    loading strategy, running backtest, recording orders, writing logs, and sending 
    notifications.

    Usually a scheduler, at time as specified in a trade plan, instantiates a `BacktestRunner`,
    and calls its `execute()` method which starts a backtest in an isolated process.
    The process runs the backtest and writes log into database.
    '''

    def __init__(self, plan: TradePlan):
        self.plan = plan
        self.run_id = None
        self.data = None
        self.strategy = None
        self.code = None
        self.strategy = None
        self.result = None
        self.params = None

    def _check_data(self, data: pd.DataFrame):
        '''Check if price data is valid.'''
        # Check if data is invariant with previous backtest data.
        ohlc = ['Open', 'High', 'Low', 'Close']
        if self.plan.strict:
            logs = self.plan.list_logs()
            log = next((l for l in logs if l.data is not None and not l.error), None)
            if log:
                # exclude the last data point which may change intraday
                # only check for price change as volume data from Yahoo do change sometimes
                prefix_len = len(log.data) - 1
                for col in ohlc:
                    if not np.allclose(
                            log.data.xs(col, 1, 1).iloc[: prefix_len],
                            data.xs(col, 1, 1).iloc[: prefix_len],
                            rtol=1e-5):
                        raise RuntimeError(
                            'Data change detected. If this is due to dividend or stock split, please start a new trade plan.')
        # Check if most recent data are actually updated, otherwise issue a warning
        if len(data) > 1:
            same = np.isclose(data.iloc[-1].loc[:, ohlc].astype(float),
                              data.iloc[-2].loc[:, ohlc].astype(float), rtol=1e-5)
            if same.any():
                print('Warning: Data may not be updated. Please verify.')
                print(data.iloc[-2:])

    def run_backtest(self, run_id: str = None, dryrun: bool = False, **kwargs: dict[str, Any]) -> pd.Series | None:
        '''Run backtest according to the trade plan and log the result to database.

        Args:
            run_id: A string ID to uniquely identify this backtest run. Leave as None to
                automatically generate one.
            dryrun: True to backtest in dryrun mode, in which case no orders are generated.
            **kwargs: Keyword parameters that are passed to Backtest.run().

        Returns:
            Backtest result as pd.Series or None if backtest is not successful
        '''

        def _record_backtest_params(**params):
            '''Record backtest run to database.'''
            bt = Backtest(**params)
            params.pop('strategy')
            params.pop('data')
            self.params = deepcopy(params)
            return bt

        def _get_trade_start_date():
            '''Trade start date is today if today is a valid trading day and it's after market open time. 
            Otherwise, it's the last valid trading day.
            '''
            calendar = mcal.get_calendar(self.plan.market_calendar)
            today = datetime.now(ZoneInfo(self.plan.market_timezone))
            valid_days = calendar.valid_days(start_date=(today - timedelta(days=30)).strftime('%Y-%m-%d'),
                                             end_date=today.strftime('%Y-%m-%d'), tz=self.plan.market_timezone)
            if today.replace(hour=0, minute=0, second=0, microsecond=0) in valid_days:
                if today.time() > calendar.open_time:
                    trade_start_date = today.strftime('%Y-%m-%d')
                else:
                    trade_start_date = valid_days[-2].strftime('%Y-%m-%d')
            else:
                trade_start_date = valid_days[-1].strftime('%Y-%m-%d')
            return trade_start_date

        exception = None
        try:
            self.run_id = run_id if run_id else MTDB.uniqueid()
            self.code = StrategyManager.read(self.plan.strategy_file)
            self.strategy = StrategyManager.load(self.plan.strategy_file)
            if self.plan.data_source:
                source = QuoteSource.get_source(self.plan.data_source)
                self.data = source.daily_bar(tickers=self.plan.ticker_css, start=self.plan.backtest_start_date)
            else:
                self.data = self.strategy.prepare_data(self.plan.ticker_css.split(','), self.plan.backtest_start_date)
            self._check_data(self.data)
            if self.plan.strict:
                bt = _record_backtest_params(strategy=self.strategy,
                                             data=self.data,
                                             cash=self.plan.initial_cash,
                                             holding=self.plan.initial_holding,
                                             commission=self.plan.commission_rate,
                                             trade_on_close=self.plan.entry_type == 'TOC',
                                             trade_start_date=self.plan.trade_start_date,
                                             **kwargs)
                self.result = bt.run()
            else:
                account = BrokerAccount.get_account(self.plan)
                broker = Broker.get_broker(account)
                # get the latest portfolio info from broker
                broker.connect()
                broker.download_orders()
                broker.download_trades()
                orders = self.plan.get_orders()
                trades = broker.format_trades(orders)
                current_holding, current_cash = calculate_positions(self.plan, trades)
                # get backtest start date
                trade_start = _get_trade_start_date()
                # cancel all raw orders not yet submitted
                self.plan.cancel_pending_orders()
                # cancel all orders submitted but not yet filled
                broker.cancel_order(self.plan)
                # run backtest
                bt = _record_backtest_params(strategy=self.strategy,
                                             data=self.data,
                                             cash=current_cash,
                                             holding=current_holding,
                                             commission=self.plan.commission_rate,
                                             trade_on_close=self.plan.entry_type == 'TOC',
                                             trade_start_date=trade_start,
                                             storage=self.plan.storage,
                                             **kwargs)
                self.result = bt.run()
            if self.result is not None and not dryrun:
                self._record_orders(ignore_run_id=self.plan.strict)
                self.plan.save_storage()
            return self.result
        except Exception:
            exception = traceback.format_exc()
        finally:
            self._log_backtest_run(exception=exception)

    def _record_orders(self, ignore_run_id) -> None:
        '''Record raw orders from a backtest run to database.

        Backtest is expected to be repeatable in strict mode, i.e., the orders generated by two backtest runs
        are exactly the same for the period when the two runs overlap. Then the backtest run on
        a later day will possibly generates some new orders based on new data not seen by the 
        earlier run. Such new orders will be recorded in associated with the later run.
        '''

        def hash(s: pd.Series):
            # include fields that can uniquely identify an order
            if ignore_run_id:
                df = s.loc[['plan_id', 'ticker', 'size', 'signal_time']]
            else:
                df = s.loc[['run_id', 'plan_id', 'ticker', 'size', 'signal_time']]
            signature = df.to_json(date_format='iso').encode('utf-8')
            return hashlib.md5(signature).hexdigest()

        # record orders from the backtest result to database
        orders: pd.DataFrame = self.result['_orders'].reset_index()
        orders.rename(columns={'SignalTime': 'signal_time', 'Ticker': 'ticker',
                               'Side': 'side', 'Size': 'size'}, inplace=True)
        orders['plan_id'] = self.plan.id
        orders['run_id'] = self.run_id
        orders['id'] = orders.apply(lambda x: hash(x), axis=1)
        orders['cancelled'] = False
        MTDB.save('RawOrder', orders.to_dict('records'), on_conflict='ignore')

    def execute(self, dryrun: bool = False) -> BacktestLog:
        '''Run backtest in an isolated process.

        Args:
            dryrun: True to run backtest in dryrun mode, in which case no orders are generated.

        Returns:
            `BacktestLog` of the run
        '''
        try:
            self.run_id = MTDB.uniqueid()
            env = os.environ.copy()
            env.update({'TQDM_DISABLE': '1'})
            proc = subprocess.run(
                [sys.executable, '-m', 'minitrade.cli', 'backtest', '--run_id', self.run_id, self.plan.id] +
                (['--dryrun'] if dryrun else []) + (['--pytest'] if 'pytest' in sys.modules else []),
                capture_output=True,
                cwd=os.getcwd(),
                timeout=3600,
                env=env,
            )
            self._log_backtest_run(stdout=proc.stdout if proc.stdout else None,
                                   stderr=proc.stderr if proc.stderr else None)
            return self.plan.get_log(self.run_id)
        except Exception as e:
            ex = traceback.format_exc()
            self._log_backtest_run(exception=ex,
                                   stdout=proc.stdout if proc.stdout else None,
                                   stderr=proc.stderr if proc.stderr else None)
            raise RuntimeError(f'Backtest for {self.plan} run {self.run_id} failed') from e
        finally:
            self._send_backtest_notification()

    def _log_backtest_run(self,
                          exception: str = None,
                          stdout: bytes | None = None,
                          stderr: bytes | None = None) -> None:
        '''Log the input and output of a backtest run to database.'''
        log = self.plan.get_log(self.run_id)
        if log is None:
            log = BacktestLog(
                id=self.run_id,
                plan_id=self.plan.id,
                plan_name=self.plan.name,
                plan_strategy=self.plan.strategy_file,
                plan=self.plan,
                data=self.data,
                strategy_code=self.code,
                params=self.params,
                result=self.result,
                exception=exception,
                stdout=stdout,
                stderr=stderr,
                log_time=datetime.utcnow()
            )
            log.save()
        else:
            MTDB.update('BacktestLog', 'id', log.id, values={
                'exception': '\n\n'.join([x for x in (log.exception, exception) if x]) or None,
                'stdout': stdout,
                'stderr': stderr,
            })

    def _send_backtest_notification(self):
        ''' Send backtest result via telegram and/or email '''
        log = self.plan.get_log(self.run_id)
        if log:
            orders = self.plan.get_orders(log.id)
            plan_name = log.plan_name
            plan_status = 'Failed' if log.error else 'Succeeded'
            plan_time = log.log_time.replace(tzinfo=ZoneInfo('UTC')).astimezone(
                ZoneInfo(self.plan.market_timezone)).strftime('%Y-%m-%d %H:%M:%S')
            plan_subject = f'## {plan_name} ##  {plan_status} @ {plan_time}'
            plan_positions = plan_data = plan_result = plan_exception = plan_stderr = plan_stdout = None

            if log.data is not None:
                data = log.data.xs('Close', 1, 1).tail(2).T
                data.columns = pd.to_datetime(data.columns).strftime('%Y-%m-%d')
                plan_data = tabulate(data.head(20), stralign="right", headers='keys')
                if len(data) > 20:
                    plan_data += '\n... (truncated)'

            if log.result is not None:
                if '_positions' in log.result.index:
                    positions = log.result.loc['_positions'][0]
                    positions = json.loads(positions.replace("'", '"'))
                    if len(positions) > 20:
                        positions = {k: v for k, v in positions.items() if v} | {'Others...': 0}
                    plan_positions = tabulate(positions.items(), headers=['Ticker', 'Share'])
                result = log.result
                result = result[~result.index.str.startswith('_')].T.apply(pd.to_numeric, errors='ignore')
                result = result.round(2).T
                for item in ['Start', 'End']:
                    result.loc[item] = result.loc[item][0][:10]
                for item in [
                    'Duration', 'Max. Drawdown Duration', 'Avg. Drawdown Duration', 'Max. Trade Duration',
                        'Avg. Trade Duration']:
                    result.loc[item] = str(result.loc['Duration'][0])[:-10]
                plan_result = tabulate(result, stralign="right")

            plan_exception = log.exception
            plan_stderr = log.stderr
            plan_stdout = log.stdout

            summary = {
                'Status': [plan_status],
                'Exception': [plan_exception is not None],
                'Stderr': [plan_stderr is not None],
                'Stdout': [plan_stdout is not None],
            }

            message = [f'<b>{plan_subject}</b>']
            message.append(f'<pre>{html.escape(tabulate(summary, headers="keys"))}</pre>')

            if orders:
                plan_orders = tabulate([[o.side, o.ticker, abs(o.size)] for o in orders])
                message.append(f'<b>Orders</b>\n<pre>{html.escape(plan_orders)}</pre>')
            else:
                message.append(f'<b>Orders</b>\n<pre>No order generated</pre>')

            if plan_positions:
                message.append(f'<b>Positions</b>\n<pre>{html.escape(plan_positions)}</pre>')

            if plan_data:
                message.append(f'<b>Data</b>\n<pre>{html.escape(plan_data)}</pre>')

            if plan_result:
                message.append(f'<b>Result</b>\n<pre>{html.escape(plan_result)}</pre>')

            if plan_exception:
                message.append(f'<b>Exception</b>\n<pre>{html.escape(plan_exception)}</pre>')

            if plan_stderr:
                message.append(f'<b>Stderr</b>\n<pre>{html.escape(plan_stderr)}</pre>')

            if plan_stdout:
                message.append(f'<b>Stdout</b>\n<pre>{html.escape(plan_stdout)}</pre>')

            message = '\n\n'.join(message)
            # send silent message if the plan runs multiple times a day and there is no error and no order generated
            silent = not log.error and not orders and len(self.plan.trade_time_of_day.split(',')) > 1
            send_telegram_message(html=message, silent=silent)
            if not silent:
                mailjet_send_email(plan_subject, message)


@dataclass(kw_only=True)
class TraderLog:
    id: str
    summary: str
    start_time: datetime
    log_time: datetime

    def save(self):
        MTDB.save('TraderLog', self, on_conflict='error')


class Trader:
    '''Trader handles order submission.

    A scheduler invokes Trader periodically, which checks for new orders generated from different 
    backtest runs, validates, and submits them to brokers.
    '''

    def submit_orders(self, plan: TradePlan):
        '''Submit orders for a particular trade plan. 

        Args:
            plan: Trade plan
        '''
        orders = MTDB.get_all(
            'RawOrder', where={'plan_id': plan.id, 'broker_order_id': None, 'cancelled': False},
            orderby='signal_time', cls=RawOrder)

        if not orders:
            return

        start_time = datetime.utcnow()
        account = BrokerAccount.get_account(plan)
        broker = Broker.get_broker(account)

        if not broker.is_ready():
            send_telegram_message(
                f'Plan {plan.name}',
                f'{len(orders)} order to be submitted:',
                *['- ' + x.tag for x in orders],
                'Use "/ib login" to login.'
            )
            return

        run_id = MTDB.uniqueid()
        summary = [f'<b>Trader ## {plan.name} ##</b><pre> </pre>']

        try:
            # download the latest orders and trades from broker before submitting new orders
            broker.download_orders()
            broker.download_trades()
            # submit orders
            status = [[o.ticker, o.side, abs(o.size), 'ABORT'] for o in orders]
            ex = None
            for i, order in enumerate(orders):
                try:
                    order.broker_order_id = broker.submit_order(plan, order)
                    order.save()
                except Exception:
                    status[i][3] = 'ERROR'
                    ex = traceback.format_exc()
                    break
                else:
                    status[i][3] = 'OK'
            summary.append(
                f'<b>Submit orders</b>\n<pre>{html.escape(tabulate(status))}</pre>')
            if ex:
                summary.append(f'<b>Exception</b>\n<pre>{html.escape(ex)}</pre>')
            # download orders and trades after submitting new orders
            broker.download_orders()
            broker.download_trades()
        except Exception:
            ex = traceback.format_exc()
            summary.append(f'<b>Exception</b>\n<pre>{html.escape(ex)}</pre>')
        finally:
            self._log_trader_run(run_id, summary, start_time)

    def execute(self):
        '''Iterate all trade plans and submit orders if there are any new ones.

        Disabled trade plans are not processed.
        '''
        for plan in TradePlan.list():
            if plan.enabled:
                self.submit_orders(plan)

    def _log_trader_run(self, run_id: str, summary: list[str], start_time: datetime) -> None:
        '''Log the input and output of a trader run to database.'''
        text = '\n\n'.join(summary)
        log = TraderLog(
            id=run_id,
            summary=text,
            start_time=start_time,
            log_time=datetime.utcnow()
        )
        log.save()
        send_telegram_message(html=text)
        mailjet_send_email(
            f'Trader @ {start_time.replace(tzinfo=ZoneInfo("UTC")).astimezone().strftime("%Y-%m-%d %H:%M:%S")}', text)


@dataclass(kw_only=True)
class TaskLog:

    id: str
    '''Task run ID'''

    plan_id: str
    '''Task plan ID'''

    plan_name: str
    '''Task plan name'''

    task_code: str
    '''Task code snapshot'''

    return_value: int | None = None
    '''Return value of the task'''

    stdout: str | None = None
    '''STDOUT capture during task execution'''

    stderr: str | None = None
    '''STDERR capture during task execution'''

    log_time: datetime
    '''Log time'''

    @property
    def error(self) -> bool:
        '''A task is regarded as failed if return value is not 0

        Returns:
            True for error, False for success
        '''
        return self.return_value != 0

    def save(self):
        MTDB.save('TaskLog', self, on_conflict='error')


@dataclass(kw_only=True)
class TaskPlan:

    id: str
    '''Task plan ID'''

    name: str
    '''A unique human readable name'''

    task_file: str
    '''Task file name'''

    timezone: str
    '''Timezone of the task plan'''

    schedule: str | None
    '''Crontabs of when the task plan should run'''

    notification: str | None
    '''Notification settings in JSON format'''

    enabled: bool = False
    '''If the task plan is enabled'''

    create_time: datetime
    '''Time when the task plan is created'''

    update_time: datetime | None
    '''Time when the task plan is updated'''

    @staticmethod
    def list() -> list[TaskPlan]:
        '''Return the list of available task plan

        Returns:
            A list of zero or more task plans
        '''
        return MTDB.get_all('TaskPlan', orderby='name', cls=TaskPlan)

    @staticmethod
    def get_plan(plan_id_or_name: str) -> TaskPlan:
        '''Look up a task by plan ID or plan name

        Args:
            plan_id_or_name: Task plan ID or name

        Returns:
            Task plan if found or None
        '''
        plan: TaskPlan = (MTDB.get_one('TaskPlan', 'id', plan_id_or_name, cls=TaskPlan)
                          or MTDB.get_one('TaskPlan', 'name', plan_id_or_name, cls=TaskPlan))
        return plan

    def __call_scheduler(self, method: str, path: str, params: dict | None = None) -> Any:
        url = f'http://{config.scheduler.host}:{config.scheduler.port}{path}'
        resp = requests.request(method=method, url=url, params=params)
        if resp.status_code == 200:
            return resp.json()
        elif resp.status_code >= 400:
            raise RuntimeError(f'Scheduler {method} {url} {params} returns {resp.status_code} {resp.text}')

    def enable(self, enable: bool):
        '''Enable or disable a plan.

        Args:
            enable: True to schedule a plan for running, or False to deschedule.
        '''
        if enable:
            if not self.schedule or not self.timezone:
                raise RuntimeError('Task plan must have schedule and timezone to be enabled')
        self.enabled = enable
        self.update_time = datetime.utcnow()
        MTDB.update('TaskPlan', 'id', self.id, values={'enabled': self.enabled, 'update_time': self.update_time})
        self.__call_scheduler('PUT', f'/task/{self.id}')

    def jobinfo(self) -> dict:
        '''Get scheduled job status of the task plan.

        Returns:
            A dict like {job_id: ..., job_frequency: ..., next_run_time: ...}
        '''
        return self.__call_scheduler('GET', f'/task/{self.id}')

    def save(self) -> None:
        '''Schedule the task plan and save it to database.

        Fail if plan ID or plan name already exists.
        '''
        MTDB.save('TaskPlan', self, on_conflict='update')
        self.__call_scheduler('PUT', f'/task/{self.id}')

    def delete(self) -> None:
        '''Unschedule a task plan and delete it from database.
        '''
        MTDB.delete('TaskPlan', 'id', self.id)
        self.__call_scheduler('DELETE', f'/task/{self.id}')

    def list_logs(self, limit=30) -> list[TaskLog]:
        '''Return all backtest history for this task plan.

        Returns:
            A list of zero or more task logs
        '''
        return MTDB.get_all(
            'TaskLog', 'plan_id', self.id, orderby=('log_time', False),
            limit=limit, cls=TaskLog)

    def get_log(self, run_id: str) -> TaskLog:
        '''Return task log for a particular run.

        Args:
            run_id: The task run ID

        Returns:
            Backtest log, or None if not found
        '''
        return MTDB.get_one('TaskLog', 'id', run_id, cls=TaskLog)


class TaskManager:
    '''TaskManager provides methods to organizes task files and their execution.'''

    __task_root = os.path.expanduser('~/.minitrade/task')

    @staticmethod
    def locate(filename: str) -> str:
        ''' Get the absolute path of a task file '''
        return os.path.join(TaskManager.__task_root, filename)

    @staticmethod
    def list() -> list[str]:
        '''Return the list of available task files

        Returns:
            A list of task filenames
        '''
        return [f for f in os.listdir(TaskManager.__task_root) if os.path.isfile(TaskManager.locate(f))]

    @staticmethod
    def save(filename: str, content: str):
        '''Save `content` to a strategy file. 

        Args:
            filename: Strategy file name
            content: File content as string
        '''
        with open(TaskManager.locate(filename), 'w') as f:
            f.write(content)
        plan = TaskPlan.get_plan(filename)
        if plan:
            plan.update_time = datetime.utcnow()
        else:
            plan = TaskPlan(id=MTDB.uniqueid(), name=filename, task_file=filename, timezone=None,
                            schedule=None, notification=None, enabled=False,
                            create_time=datetime.utcnow(), update_time=None)
            with open(TaskManager.locate(filename), 'r') as f:
                line = f.readline()
            try:
                # The first line in file can be a comment that contains default config like:
                # # Default config: {"timezone": "America/New_York", "schedule": "0 9 * * MON-FRI", "enabled": true, "notification": {"telegram": "N", "email": "N"}}
                if line.startswith('# Default config:'):
                    config = json.loads(line[17:])
                    timezone = config.get('timezone')
                    schedule = config.get('schedule')
                    enabled = config.get('enabled', False)
                    notification = config.get('notification')
                    plan = TaskPlan(id=MTDB.uniqueid(), name=filename, task_file=filename, timezone=timezone,
                                    schedule=schedule, notification=notification, enabled=enabled,
                                    create_time=datetime.utcnow(), update_time=None)
            except Exception:
                logger.warning(f'Ignore invalid default config in "{filename}": {line}')
        plan.save()

    @staticmethod
    def read(task_file: str) -> str:
        ''' Return the content of `task_file` 

        Args:
            task_file: The task file name

        Returns:
            The file content as a string
        '''
        with open(TaskManager.locate(task_file), 'r') as f:
            return f.read()

    @staticmethod
    def delete(task_file: str) -> None:
        '''Delete `strategy_file`

        Args:
            task_file: The task file name
        '''
        os.remove(TaskManager.locate(task_file))


class TaskRunner:
    '''TaskRunner handles the execution of a task, write logs, and send notifications.'''

    def __init__(self, plan: TaskPlan):
        self.plan = plan
        self.code = TaskManager.read(self.plan.task_file)

    def execute(self) -> TaskLog:
        '''Run task in an isolated process.

        Returns:
            `TaskLog` of the run
        '''
        try:
            proc = subprocess.run(
                [sys.executable, TaskManager.locate(self.plan.task_file)],
                capture_output=True,
                cwd=TaskManager.locate('.'),
                timeout=3600)
            log = TaskLog(
                id=MTDB.uniqueid(),
                plan_id=self.plan.id,
                plan_name=self.plan.name,
                task_code=self.code,
                return_value=proc.returncode,
                stdout=proc.stdout if proc.stdout else None,
                stderr=proc.stderr if proc.stderr else None,
                log_time=datetime.utcnow()
            )
            log.save()
        except Exception:
            ex = traceback.format_exc()
            log = TaskLog(
                id=MTDB.uniqueid(),
                plan_id=self.plan.id,
                plan_name=self.plan.name,
                task_code=self.code,
                return_value=10000,
                stdout=None,
                stderr=ex,
                log_time=datetime.utcnow()
            )
            log.save()
        finally:
            log = self.plan.get_log(log.id)
            self._send_task_notification(log)
            return log

    def _send_task_notification(self, log: TaskLog):
        ''' Send task result via telegram and/or email '''
        telegram = self.plan.notification.get('telegram', 'N') if self.plan.notification else 'N'
        email = self.plan.notification.get('email', 'N') if self.plan.notification else 'N'
        subject = f'Task "{self.plan.name}" {"❌" if log.error else "✅"}  @ {log.log_time.strftime("%Y-%m-%d %H:%M:%S")}'
        message = [f'<b>{subject}</b>\n<pre></pre>']
        message.append(f'<b>Return value</b>\n<pre>{html.escape(str(log.return_value))}</pre>')
        message.append(f'<b>STDERR</b>\n<pre>{html.escape(log.stderr or "None")}</pre>')
        message.append(f'<b>STDOUT</b>\n<pre>{html.escape(log.stdout or "None")}</pre>')
        message = '\n\n'.join(message)
        if telegram == 'A' or (telegram == 'E' and log.error):
            send_telegram_message(html=message)
        if email == 'A' or (email == 'E' and log.return_value != 0):
            mailjet_send_email(subject, message)





from datetime import datetime, timedelta

from alpaca.data import (Adjustment, CryptoHistoricalDataClient, Sort,
                         StockBarsRequest, StockHistoricalDataClient,
                         TimeFrame, TimeFrameUnit)

from minitrade.datasource.base import QuoteSource
from minitrade.utils.config import config


class AlpacaQuoteSource(QuoteSource):
    '''AlpacaQuoteSource retrieves data from https://alpaca.markets. (Experimental)
    '''

    def __init__(self, api_key: str = None, api_secret: str = None, use_adjusted: bool = True) -> None:
        super().__init__()
        api_key = api_key or config.sources.alpaca.api_key
        api_secret = api_secret or config.sources.alpaca.api_secret
        self.use_adjusted = use_adjusted
        if not api_key or not api_secret:
            raise AttributeError('Alpaca API key or secret is not configured')
        self.crypto_client = CryptoHistoricalDataClient()
        self.stock_client = StockHistoricalDataClient(api_key, api_secret)

    def _format_ticker(self, ticker):
        return ticker

    def _ticker_timezone(self, ticker):
        # TODO: get timezone from API
        return 'America/New_York'

    def _ticker_calendar(self, ticker):
        # TODO: get market calendar from API
        return 'NYSE'

    def _daily_bar(self, ticker, start, end):
        request = StockBarsRequest(
            symbol_or_symbols=ticker,
            timeframe=TimeFrame(1, TimeFrameUnit.Day),
            start=datetime.strptime(start, '%Y-%m-%d'),
            end=datetime.strptime(end, '%Y-%m-%d') + timedelta(hours=12) if end else None,
            adjustment=Adjustment.ALL if self.use_adjusted else Adjustment.NONE,
            sort=Sort.ASC)
        df = self.stock_client.get_stock_bars(request_params=request).df.loc[ticker]
        df.rename(columns={'open': 'Open', 'high': 'High', 'low': 'Low',
                  'close': 'Close', 'volume': 'Volume'}, inplace=True)
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        df.index.rename('Date', inplace=True)
        df.index = df.index.tz_localize(None).normalize()
        return df

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        raise NotImplementedError()

    def _spot(self, tickers):
        raise NotImplementedError()




import zoneinfo
from dataclasses import asdict
from datetime import datetime, time, timedelta

import pandas as pd
import pandas_market_calendars as mcal
import streamlit as st
from matplotlib import pyplot as plt
from tabulate import tabulate

from minitrade.backtest import calculate_trade_stats
from minitrade.broker import Broker, BrokerAccount
from minitrade.datasource import QuoteSource
from minitrade.trader import (BacktestLog, BacktestRunner, StrategyManager,
                              TradePlan)
from minitrade.trader.scheduler import run_trader_after_backtest
from minitrade.utils.mtdb import MTDB

st.set_page_config(page_title='Trading', layout='wide')


def get_market_calendar_and_timezone(data_source: str, ticker_css: str):
    ''' Get market calendar and timezone from tickers '''
    if ticker_css.strip() == '':
        return None, None
    ds = QuoteSource.get_source(data_source)
    calendars = list(dict.fromkeys([ds.ticker_calendar(t) for t in ticker_css.split(',')]))
    zones = list(dict.fromkeys([ds.ticker_timezone(t) for t in ticker_css.split(',')]))
    if len(calendars) > 1 or len(zones) > 1:
        st.info(
            f'Cross market trading detected. Use "{calendars[0]}" for market calendar, "{zones[0]}" for market timezone.')
    return calendars[0], zones[0]


def ticker_resolver(account: BrokerAccount, ticker_css: str) -> dict:
    ''' Map ticker to broker specific contract ID '''
    if account and len(ticker_css.replace(' ', '')) > 0:
        broker = Broker.get_broker(account)
        with st.expander('Resolve tickers', expanded=True):
            try:
                broker.connect()
                candidates: dict = broker.resolve_tickers(ticker_css.replace(' ', ''))
                if candidates:
                    vendor_tickers = {}
                    for k, v in candidates.items():
                        vendor_tickers[k] = st.radio(k, v, format_func=lambda _: _['label'])
                    return vendor_tickers
            except ConnectionError:
                st.error('Need to login to broker account to resolve tickers. Please pay attention to 2FA notification if enabled.')
                st.button('Retry')
    return {}


def parse_trade_time_of_day(trade_time_of_day: str) -> str:
    ''' Parse trade time of day from string '''
    if trade_time_of_day:
        ts = []
        for t in trade_time_of_day.split(','):
            t = t.strip()
            if len(t) > 8:
                start_end, interval = t.split('/')
                start, end = start_end.split('-')
                ts.extend(pd.date_range(start, end, freq=interval).time)
            else:
                ts.append(datetime.strptime(t, '%H:%M:%S').time() if len(
                    t.split(':')) == 3 else datetime.strptime(t, '%H:%M').time())
        return ','.join([t.strftime('%H:%M:%S') for t in sorted(ts)])
    else:
        return None


def show_create_trade_plan_form() -> TradePlan | None:
    account = st.selectbox('Select a broker account', BrokerAccount.list(), format_func=lambda b: b.alias)
    strategy_file = st.selectbox('Pick a strategy', StrategyManager.list())
    data_source = st.selectbox(
        'Select a data source', QuoteSource.list() + [None], index=QuoteSource.list().index('Yahoo'),
        format_func=lambda x: x or 'Use Strategy.prepare_data()')
    ticker_css = st.text_input(
        'Define the asset space (a list of tickers separated by comma without space)', placeholder='e.g. AAPL,GOOG')
    market_calendar, market_timezone = None, None
    try:
        if data_source:
            market_calendar, market_timezone = get_market_calendar_and_timezone(data_source, ticker_css)
    except Exception as e:
        st.warning(f'{e}. Please select market and timezone manually.')
    c1, c2 = st.columns(2)
    markets = mcal.get_calendar_names()
    zones = sorted(list(zoneinfo.available_timezones()))
    market_calendar = c1.selectbox('Select market', options=markets, index=markets.index(market_calendar or 'NYSE'))
    market_timezone = c2.selectbox('Select market timezone', options=zones, index=zones.index(
        market_timezone or 'America/New_York'))
    backtest_start_date = st.date_input(
        'Pick a backtest start date (run backtest from that date to give enough lead time to calculate indicators)',
        value=datetime.today() - timedelta(days=30))
    trade_start_date = st.date_input(
        'Pick a trade start date (trade signal before that is surpressed)', min_value=datetime.now().date())
    c1, c2 = st.columns([1, 3])
    entry_type = c1.selectbox('Select order entry type', ['TOO', 'TOC', 'TRG'], format_func=lambda x: {
        'TOO': 'Trade on open (TOO)', 'TOC': 'Trade on close (TOC)', 'TRG': 'Trade regular hours (TRG)'}[x])
    trade_time_of_day = c2.text_input(
        'Set when backtest should run (in market local timezone as a list of HH\\:MM[\\:SS] separated by comma)',
        placeholder='e.g. 9:30, 15:59:30, 9:30:01-16:00/30min')
    try:
        trade_time_of_day = parse_trade_time_of_day(trade_time_of_day)
        if trade_time_of_day:
            st.info(f'{len(trade_time_of_day.split(","))} runs @ {trade_time_of_day.replace(",", ", ")}')
    except Exception as e:
        st.error(e)
    initial_cash = st.number_input('Set the cash amount to invest', value=0)
    initial_holding = st.text_input(
        'Set the preexisting asset positions (number of shares you already own and to be considered in the strategy)',
        placeholder='e.g. AAPL:100,GOOG:100') or None
    name = st.text_input('Name the trade plan')
    strict = st.radio('Select backtest mode', ['Strict', 'Incremental'], index=0) == 'Strict'
    st.info(
        'In strict mode, repeated backtests are expected to produce identical results, and orders '
        'generated from backtests are expected to be executed successfully at all time. This attempts '
        'to trade a strategy as faithfully as possible. However, it may suffer from multiple external '
        'changes, e.g. data change due to dividend and stock split, failed order execution, etc., '
        'in which case, trading will fail to proceed, and manual intervention is required. '
        'In incremental mode, backtests will be run incrementally on the latest data only and in observation of the '
        'actual portfolio status. Strategies must be designed to run incrementally, i.e. not dependent on previous runs. '
        'This mode is more robust to external changes, but performance may deviate from that of the original strategy. '
        'See [Trading](https://dodid.github.io/minitrade/trading/) for more details.')

    tickers = ticker_resolver(account, ticker_css)
    ticker_map = {k: v['id'] if v is not None else None for k, v in tickers.items()}
    if account and None in ticker_map.values():
        st.warning('Tickers are not fully resolved. You can ignore this warning if unresolved tickers are not traded.')
    dryrun = st.button('Save and dry run')

    if dryrun:
        if initial_holding:
            initial_holding = {k.strip(): int(v)
                               for k, v in [x.strip().split(':') for x in initial_holding.split(',')]}
        if len(name.strip()) == 0:
            st.error('Trade plan name is required')
        elif not trade_time_of_day:
            st.error('Trade time of day is required')
        else:
            return TradePlan(
                id=MTDB.uniqueid(),
                name=name,
                strategy_file=strategy_file,
                ticker_css=ticker_css.replace(' ', ''),
                market_calendar=market_calendar,
                market_timezone=market_timezone,
                data_source=data_source,
                backtest_start_date=backtest_start_date.strftime('%Y-%m-%d'),
                trade_start_date=trade_start_date.strftime('%Y-%m-%d'),
                trade_time_of_day=trade_time_of_day,
                entry_type=entry_type,
                broker_account=account.alias if account else None,
                initial_cash=initial_cash,
                initial_holding=initial_holding,
                strict=strict,
                enabled=False,
                create_time=datetime.utcnow(),
                update_time=None,
                broker_ticker_map=ticker_map
            )


def display_run(plan: TradePlan, log: BacktestLog):
    orders = plan.get_orders(log.id)
    log_status = '❌' if log.error else '✅' if orders else '🟢'
    log_time = log.log_time.replace(tzinfo=zoneinfo.ZoneInfo('UTC')).astimezone(
        zoneinfo.ZoneInfo(plan.market_timezone)).strftime('%Y-%m-%d %H:%M:%S')
    label = f'{log_status} {log_time}' + (f' **{len(orders)} orders**' if orders else '')
    with st.expander(label):
        tab1, tab2, tab3, tab4, tab5, tab6 = st.tabs(
            ['Input', 'Data', 'Result', 'Error', 'Log', 'Orders'])
        if log.params:
            tab1.write(log.params)
        if log.data.size > 100_000:
            tab2.write(f'Data too large to display, shape: {log.data.shape}')
        else:
            tab2.write(log.data)
        if log.result is not None:
            df = log.result
            df = df[~df.index.str.startswith('_')]
            tab3.text(tabulate(df))
        tab4.code(log.exception)
        tab5.caption('Log - stdout')
        if log.stdout:
            tab5.text(log.stdout)
        tab5.caption('Log - stderr')
        if log.stderr:
            tab5.text(log.stderr)
        if orders:
            tab6.write(pd.DataFrame(orders))


def save_plan_and_dryrun(plan: TradePlan) -> None:
    plan.save()
    st.success('Trade plan saved')
    runner = BacktestRunner(plan)
    log = runner.execute(dryrun=True)
    display_run(plan, log)
    if log.error:
        st.error('Trade plan dryrun failed, plan disabled')
    else:
        plan.enable(True)
        st.success('Trade plan dryrun succeeded, plan enabled')


def show_trade_plan_selector() -> TradePlan | None:
    plan_lst = TradePlan.list()
    plan_idx = st.sidebar.radio('Trade plan', list(range(len(plan_lst))),
                                format_func=lambda i: plan_lst[i].name)
    plan = plan_lst[int(plan_idx)] if plan_idx is not None else None
    return plan


def run_trade_plan_once(plan: TradePlan) -> None:
    log = run_trader_after_backtest(plan, force=True)
    if log is not None and not log.error:
        st.success(f'Backtest {log.id} finished successfully')
    else:
        st.error(f'Backtest failed')


def confirm_delete_trade_plan(plan: TradePlan) -> None:
    def confirm_delete():
        if st.session_state.delete_confirm_textinput == plan.name:
            plan.delete()
    st.text_input(f'Type "{plan.name}" and press Enter to delete',
                  on_change=confirm_delete, key='delete_confirm_textinput')


def show_trade_plan_header_and_controls(plan: TradePlan) -> None:
    c1, c2, c3, c4, c5 = st.columns([5, 1, 1, 1, 1])
    c1.subheader(plan.name)
    c2.button('Refresh')
    c3.button('Run once', on_click=lambda: run_trade_plan_once(plan))
    c4.button('Disable' if plan.enabled else 'Enable', key='tradeplan_onoff',
              on_click=lambda: plan.enable(not plan.enabled))
    if c5.button('Delete', type='primary'):
        confirm_delete_trade_plan(plan)


def show_trade_plan_status(plan: TradePlan) -> None:
    job = plan.jobinfo()
    if job is None:
        if plan.enabled:
            st.warning('State inconsistency detected: plan is enabled but not scheduled')
        else:
            st.warning('Plan is disabled and not trading')
    else:
        if plan.enabled:
            st.success(f'Plan is scheduled to run at {job["next_run_time"]}')
        else:
            st.success(
                f'State inconsistenccy detected: plan is disabled but is scheduled to run at {job["next_run_time"]}')


def show_trade_plan_execution_history(plan: TradePlan) -> None:
    tab1, tab2, tab3, tab4 = st.tabs(['Run history', 'Orders', 'Performance', 'Settings'])
    logs = plan.list_logs()
    with tab1:
        show_trade_plan_status(plan)
        for log in logs:
            display_run(plan, log)

    with tab2:
        account = BrokerAccount.get_account(plan)
        broker = account and Broker.get_broker(account)
        orders = plan.get_orders()
        show_orders(plan, account, broker, orders)

    with tab3:
        if broker and logs:
            show_performance(plan, logs, broker, orders)

    with tab4:
        st.write(asdict(plan))


def show_orders(plan, account, broker, orders):
    for order in orders:
        trade = broker and broker.find_trades(order)
        broker_order = broker and broker.find_order(order)
        order_status = '✅' if trade else '🟢' if order.broker_order_id else '🚫'
        with st.expander(f'{order_status} {order.signal_time} **{order.ticker} {order.side} {abs(order.size)}**'):
            t1, t2, t3 = st.tabs(['Raw order', 'Broker order', 'Trade status'])
            t1.dataframe(pd.DataFrame([asdict(order)]))
            if broker_order:
                t2.dataframe(pd.DataFrame([broker_order]))
            if trade:
                t3.dataframe(pd.DataFrame(trade))
                if account.broker == 'Manual':
                    c1, c2, c3, c4 = t3.columns(4)
                    price = c1.number_input('Price', key=f'{order.id}-price', value=trade[0]['price'] or 0.0)
                    commission = c2.number_input(
                        'Commission', key=f'{order.id}-comm', value=trade[0]['commission'] or 0.0)
                    dt = c3.date_input('Trade time (market local timezone)',
                                       key=f'{order.id}-date', value=trade[0]['trade_time'])
                    tm = c4.time_input('Trade time', key=f'{order.id}-time',
                                       value=trade[0]['trade_time'] or time(9, 30), label_visibility='hidden')
                    if t3.button('Save', key=f'{order.id}-save'):
                        broker.update_trade(trade[0]['id'], price, commission,
                                            datetime.combine(dt, tm, zoneinfo.ZoneInfo(plan.market_timezone)))
                        st.warning('Click "Refresh" button to see change')


def show_performance(plan, logs, broker, orders):
    try:
        data = logs[0].data
        trades = broker.format_trades(orders)
        trade_start_date = datetime.strptime(
            plan.trade_start_date, '%Y-%m-%d').replace(tzinfo=zoneinfo.ZoneInfo(plan.market_timezone))
        offset = (data.index < trade_start_date).sum()
        # calculate trade stats from trade_start_date
        # data.index may have mixed timezone due to daylight saving time
        trade_df, equity, pnl, commission_rate, cash_available = calculate_trade_stats(
            data.iloc[offset:], plan.initial_cash, trades, plan.initial_holding)
        if len(equity) > 0:
            rr = equity.pct_change()
            sharpe = rr.mean() / rr.std() * 252**0.5
            cum_rr = (equity / equity.iloc[0] - 1) * 100
            cum_rr.name = 'Return rate (%)'
            c1, c2, c3, c4 = st.columns(4)
            c1.metric(
                label='Portfolio Value', value=f'{equity.iloc[-1]:,.0f}',
                delta=f'{int(equity.iloc[-1]-equity.iloc[-2]):,}' if len(equity) >= 2 else 0)
            c2.metric(label='PnL', value=f'{(equity.iloc[-1]-equity.iloc[0]):,.0f}',
                      delta=f'{cum_rr.iloc[-1]:.2f}%')
            c3.metric(label='Sharpe Ratio', value=f'{sharpe:.2f}')
            c4.metric(label='Commission Rate', value=f'{commission_rate:.3%}')
            if len(cum_rr) >= 2:
                st.caption('Return Rate (%)')
                st.line_chart(cum_rr, height=200)
            st.caption('Profit and Loss')
            st.write(pnl.style.format('{:,.0f}', na_rep=' '))
            st.caption('Cash Available')
            st.write(f'{int(cash_available):,}')
            st.caption('Trades')
            st.write(trade_df.style.format(na_rep=' '))
            st.caption('Orders')
            broker_orders = pd.DataFrame(trades)
            broker_orders['entry_time'] = broker_orders['entry_time'].dt.tz_convert(plan.market_timezone)
            broker_orders['value'] = (broker_orders['entry_price'] * broker_orders['size']).astype(int)
            st.write(broker_orders)
        else:
            st.write('No stats due to insufficient data')
    except Exception as e:
        st.write('No stats due to insufficient data')


action = st.sidebar.radio('Action', ['Browse existing', 'Create new'])

if action == 'Create new':
    st.subheader('Create new trade plan')
    plan = show_create_trade_plan_form()
    if plan:
        save_plan_and_dryrun(plan)

if action == 'Browse existing':
    plan = show_trade_plan_selector()
    if plan:
        show_trade_plan_header_and_controls(plan)
        show_trade_plan_execution_history(plan)





import urllib.parse

import pandas as pd

from minitrade.datasource.base import QuoteSource
from minitrade.utils.config import config


class TiingoQuoteSource(QuoteSource):
    '''TiingoQuoteSource retrieves data from https://www.tiingo.com. (Experimental)
    '''

    def __init__(self, api_key: str = None, use_adjusted: bool = True) -> None:
        super().__init__()
        self.api_key = api_key or config.sources.tiingo.api_key
        self.use_adjusted = use_adjusted
        if not self.api_key:
            raise AttributeError('Tiingo API key is not configured')

    def _format_ticker(self, ticker):
        return ticker

    def _ticker_timezone(self, ticker):
        # TODO: get timezone from API
        return 'America/New_York'

    def _ticker_calendar(self, ticker):
        # TODO: get market calendar from API
        return 'NYSE'

    def _daily_bar(self, ticker, start, end):
        url = f'https://api.tiingo.com/tiingo/daily/{urllib.parse.quote_plus(ticker)}/prices?startDate={start}&format=json&resampleFreq=daily&sort=date&token={self.api_key}'
        if end:
            url += f'&endDate={end}'
        df = pd.read_json(url)
        df.set_index('date', inplace=True)
        df.index = pd.to_datetime(df.index)
        if self.use_adjusted:
            df = df[['adjOpen', 'adjHigh', 'adjLow', 'adjClose', 'adjVolume']]
            df.rename(columns={'adjOpen': 'Open', 'adjHigh': 'High', 'adjLow': 'Low',
                               'adjClose': 'Close', 'adjVolume': 'Volume'}, inplace=True)
        else:
            df = df[['open', 'high', 'low', 'close', 'volume']]
            df.rename(columns={'open': 'Open', 'high': 'High', 'low': 'Low',
                               'close': 'Close', 'volume': 'Volume'}, inplace=True)
        df.index.rename('Date', inplace=True)
        df.index = df.index.tz_localize(None).normalize()
        return df

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        raise NotImplementedError()

    def _spot(self, tickers):
        raise NotImplementedError()



import hashlib
import json
import os
import pkgutil
import platform
import sqlite3
import subprocess
import sys
import time
from contextlib import closing
from multiprocessing import Process
from pathlib import Path
from zipfile import ZipFile

import click
import requests
import yaml
from tqdm import tqdm


def __call_ibgateway_admin(method: str, path: str, params: dict | None = None):
    '''Call the ibgateway's admin API

    Parameters
    ----------
    method : str
        The HTTP method, i.e. GET, POST, PUT, DELETE, etc.
    path : str
        The REST API endpoint, e.g. '/ibgateway'
    params : dict, optional
        Extra parameters to be sent along with the REST API call

    Returns
    -------
    Any
        result json if API returns 200, otherwise None
    '''
    from minitrade.utils.config import config
    url = f'http://{config.brokers.ib.gateway_admin_host}:{config.brokers.ib.gateway_admin_port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()
    elif resp.status_code >= 400:
        raise RuntimeError(f'Request {path} returned {resp.status_code} {resp.text}')


@click.group()
@click.pass_context
def mtcli(ctx):
    pass


@mtcli.group()
def ib():
    '''Manage the execution of IB gateway'''
    pass


def ib_start_inner(log_level=None):
    import uvicorn

    from minitrade.utils.config import config
    uvicorn.run(
        'minitrade.broker.ibgateway:app',
        host=config.brokers.ib.gateway_admin_host,
        port=config.brokers.ib.gateway_admin_port,
        log_level=log_level or config.brokers.ib.gateway_admin_log_level,
    )


@ib.command('start')
def ib_start():
    '''Start IB gateway'''
    ib_start_inner()


@ib.command('stop')
def ib_stop():
    '''Stop IB gateway'''
    __call_ibgateway_admin('DELETE', '/')


@ib.command('status')
def ib_status():
    '''Check IB gateway status'''
    status = __call_ibgateway_admin('GET', '/ibgateway')
    click.echo(json.dumps(status, indent=2))


@ib.command('login')
@click.option('-a', '--alias', prompt='Alias', help='IB account alias')
def ib_login(alias):
    '''Login IB gateway to a particular account'''
    status = __call_ibgateway_admin('PUT', f'/ibgateway/{alias}')
    click.echo(status)


@ib.command('logout')
def ib_logout():
    '''Logout IB gateway from all accounts'''
    accounts = __call_ibgateway_admin('GET', '/ibgateway')
    for a in accounts:
        click.echo(a)
        __call_ibgateway_admin('DELETE', f'/ibgateway/{a["account"]}')


@mtcli.group()
def web():
    '''Manage web UI'''
    pass


@web.command('start')
def web_start():
    '''Launch web UI'''
    from streamlit.web.cli import main
    cd = os.path.dirname(os.path.realpath(__file__))
    home = os.path.join(cd, 'admin/Home.py')
    lib_home = os.path.join(cd, '/..')
    os.chdir(lib_home)
    sys.argv = ["streamlit", "run", home]
    sys.exit(main())


@mtcli.group()
def scheduler():
    '''Manage scheduler and scheduled jobs'''
    pass


def __call_scheduler(method: str, path: str, params: dict | None = None):
    '''Call the scheduler's REST API

    Parameters
    ----------
    method : str
        The HTTP method, i.e. GET, POST, PUT, DELETE, etc.
    path : str
        The REST API endpoint
    params : dict, optional
        Extra parameters to be sent along with the REST API call

    Returns
    -------
    Any
        result json if API returns 200, otherwise None
    '''
    from minitrade.utils.config import config
    url = f'http://{config.scheduler.host}:{config.scheduler.port}{path}'
    resp = requests.request(method=method, url=url, params=params)
    if resp.status_code == 200:
        return resp.json()


@scheduler.command('start')
def scheduler_start():
    '''Start scheduler'''
    import uvicorn

    from minitrade.utils.config import config
    uvicorn.run(
        'minitrade.trader.scheduler:app',
        host=config.scheduler.host,
        port=config.scheduler.port,
        log_level=config.scheduler.log_level,
    )


@scheduler.command('stop')
def scheduler_stop():
    '''Stop scheduler'''
    __call_scheduler('DELETE', '/')


@scheduler.command('status')
def scheduler_status():
    '''Check scheduler status'''
    status = __call_scheduler('GET', '/strategy')
    click.echo(json.dumps(status, indent=2))


@scheduler.command('schedule')
@click.argument('plan_id')
def scheduler_schedule(plan_id):
    '''Schedule a trade plan'''
    status = __call_scheduler('PUT', f'/strategy/{plan_id}')
    click.echo(status)


@scheduler.command('unschedule')
@click.argument('plan_id')
def scheduler_unschedule(plan_id):
    '''Unschedule a trade plan'''
    status = __call_scheduler('DELETE', f'/strategy/{plan_id}')
    click.echo(status)


@mtcli.command
@click.argument('plan_id_or_name')
@click.option('--run_id', default=None, help='Specify an unique run ID manually')
@click.option('--dryrun', is_flag=True, help='Dry run only, don\'t place orders')
@click.option('--pytest', is_flag=True, help='Run in test mode')
def backtest(plan_id_or_name, run_id, dryrun, pytest):
    '''Run backtest for particular plan'''
    import warnings
    warnings.simplefilter(action='ignore', category=FutureWarning)

    from minitrade.trader.trader import BacktestRunner
    if pytest:
        # import pytest which will trigger running in test mode using test db
        import pytest

    from minitrade.trader import TradePlan
    try:
        plan = TradePlan.get_plan(plan_id_or_name)
        runner = BacktestRunner(plan)
        runner.run_backtest(run_id=run_id, dryrun=dryrun)
    except Exception as e:
        raise e


def check_program_version(name, path=None):
    try:
        if not path:
            proc = subprocess.run(['which', name], capture_output=True, cwd=os.getcwd(), text=True)
            if proc.returncode:
                raise RuntimeError(f'Program {name} not found')
            path = proc.stdout.strip('\n')
        proc = subprocess.run([path, '--version'], capture_output=True, cwd=os.getcwd(), text=True)
        if proc.returncode:
            raise RuntimeError(f'Program {name} not found')
        version = proc.stdout.split('\n')[0]
        click.secho(f'  {name:15s} ... found {version}', fg='green')
        return True
    except Exception as e:
        click.secho(f'  {name:15s} ... missing', fg='red')
        return False


def check_selenium():
    try:
        from selenium import webdriver
        options = webdriver.ChromeOptions()
        options.add_argument('ignore-certificate-errors')
        options.add_argument("--headless")
        options.add_argument("--no-sandbox")
        options.add_argument("--disable-gpu")
        with webdriver.Chrome(options=options) as driver:
            driver.get('https://apple.com')
            if driver.find_element(value='globalnav-list'):
                click.secho(f'  {"selenium":15s} ... working', fg='green')
                return True
    except Exception as e:
        click.secho(f'  selenium ... error: {e}', fg='red')
        return False


@ib.command('diagnose')
def ib_diagnose():
    '''Diagnose IB login issues'''
    from minitrade.utils.config import config

    click.secho('Checking dependencies:')
    check_program_version('java')
    if platform.system() == 'Darwin':
        check_program_version(
            'google-chrome', '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome')
        check_program_version('chromedriver')
    elif platform.system() == 'Linux':
        check_program_version('google-chrome')

    click.secho('Checking Minitrade installation:')
    items = [
        '~/.minitrade',
        '~/.minitrade/strategy',
        '~/.minitrade/storage',
        '~/.minitrade/ibgateway',
        '~/.minitrade/config.yaml',
        '~/.minitrade/database/minitrade.db',
    ]
    for item in items:
        path = Path(item).expanduser()
        click.secho(f'  {item:35s} ... {"found" if path.exists() else "missing"}',
                    fg='green' if path.exists() else 'red')

    click.secho('Checking Selenium:')
    check_selenium()

    click.secho('Checking IB gateway:')
    try:
        if requests.get(f'http://localhost:{config.scheduler.port}/strategy', timeout=3).status_code != 200:
            raise RuntimeError('Scheduler is not running')
    except Exception:
        click.secho('  Scheduler is not running. Start it and try again.', fg='red')
        return
    try:
        if requests.get(
                f'http://localhost:{config.brokers.ib.gateway_admin_port}/ibgateway', timeout=3).status_code == 200:
            click.secho('  IB gateway is already running. Quit it and try again.', fg='red')
            return
    except Exception:
        pass
    try:
        gateway = Process(target=ib_start_inner, args=['debug'])
        gateway.start()
        time.sleep(3)
        alias = input('Please enter the IB account alias to be diagnosed: ')
        res = requests.put(f'http://localhost:{config.brokers.ib.gateway_admin_port}/ibgateway/{alias}')
        click.secho(f'Login status: {res.text}', fg='blue')
        if res.status_code != 200:
            raise RuntimeError('IB gateway login failed')
        gateway.terminate()
        time.sleep(1)
        click.secho(f'  {"ibgateway":15s} ... working', fg='green')
    except Exception as e:
        gateway.terminate()
        time.sleep(1)
        click.secho(f'  {"ibgateway":15s} ... error, {e}', fg='red')


@mtcli.command()
@click.option('-y', '--yes', is_flag=True, help='Don\'t ask for confirmation of inplace upgrade')
def init(yes):
    '''Initialize Minitrade'''
    from importlib.metadata import version

    from minitrade.utils.config import minitrade_root

    check_prerequisites()

    if os.path.exists(minitrade_root) and not yes:
        click.secho(
            f'Minitrade is already installed in {minitrade_root}. '
            f'This will update the installation to version {version("minitrade")}.\n'
            'While the operation is safe, it is recommended to backup the folder first.',
            fg='green')
        click.confirm('Do you want to continue?', abort=True)

    # create dirs
    click.secho(f'Setting up directories...')
    for d in ['database', 'strategy', 'storage', 'ibgateway', 'task']:
        os.makedirs(os.path.join(minitrade_root, d), mode=0o700, exist_ok=True)

    # upgrade config
    from minitrade.utils.config import GlobalConfig
    GlobalConfig.upgrade()

    # build db schema
    build_db_schema(minitrade_root)

    # download and extract IB gateway
    install_ibgateway(minitrade_root)

    # finish
    click.secho(f'Minitrade installed in {minitrade_root}', fg='green')


def build_db_schema(minitrade_root):
    click.secho(f'Setting up database...')
    db_loc = os.path.join(minitrade_root, 'database/minitrade.db')
    sqlfiles = [f for f in os.listdir(os.path.join(os.path.dirname(__file__), 'trader/schema')) if f.endswith('.sql')]
    sqlversions = [tuple(int(_) for _ in f.removesuffix('.sql').split('.'))
                   for f in sqlfiles if f != 'consolidated.sql']
    # read schema version from db
    try:
        with closing(sqlite3.connect(db_loc)) as conn:
            version = conn.execute('SELECT version FROM schema_version').fetchone()[0]
            version = tuple(int(_) for _ in version.split('.'))
    except Exception:
        with closing(sqlite3.connect(db_loc)) as conn:
            conn.executescript('BEGIN TRANSACTION;'
                         'CREATE TABLE IF NOT EXISTS "schema_version" ("version" TEXT);'
                         'INSERT INTO "schema_version" VALUES ("0.0.0");'
                         'COMMIT;')
        version = (0, 0, 0)
    # apply schema patches
    for sqlversion in sorted(sqlversions):
        if sqlversion <= version:
            continue
        sqlversion = '.'.join(str(_) for _ in sqlversion)
        sql = pkgutil.get_data(__name__, f'trader/schema/{sqlversion}.sql').decode('utf-8')
        sql = 'BEGIN TRANSACTION;' + sql + f'UPDATE schema_version SET version="{sqlversion}";' + 'COMMIT;'
        with closing(sqlite3.connect(db_loc)) as conn:
            conn.executescript(sql)


def install_ibgateway(minitrade_root):
    try:
        click.secho(f'Installing Interactive Brokers gateway...')
        ib_loc = os.path.join(minitrade_root, 'ibgateway')
        url = 'https://download2.interactivebrokers.com/portal/clientportal.gw.zip'
        response = requests.get(url, stream=True)
        total_kb = int(int(response.headers["Content-Length"]) / 1000)
        with open(f'{ib_loc}/clientportal.gw.zip', "wb") as f:
            file_hash = hashlib.md5()
            t = tqdm(response.iter_content(chunk_size=1000), total=total_kb,
                    unit="KB", desc='Downloading IB gateway', leave=False)
            for data in t:
                file_hash.update(data)
                f.write(data)
                t.format_sizeof(len(data))
            t.close()
            with open(f'{ib_loc}/clientportal.gw.zip.md5', "w") as h:
                h.write(file_hash.hexdigest())
        ZipFile(f'{ib_loc}/clientportal.gw.zip').extractall(ib_loc)
        # tighten API access to localhost only
        conf_loc = os.path.join(ib_loc, 'root/conf.yaml')
        with open(conf_loc, 'r') as f:
            conf = yaml.safe_load(f)
        conf['ips']['allow'] = ['127.0.0.1']
        conf['ips']['deny'] = []
        with open(conf_loc, 'w') as f:
            yaml.safe_dump(conf, f)
    except Exception as e:
        click.secho('We have some issue with installing IB gateway. If you use Minitrade solely for backtesting, you can safely ignore this.', fg='red')


def check_prerequisites():
    if platform.system() == 'Linux':
        click.secho('Checking prerequisites:')
        prerequisites = [
            check_program_version('java'),
            check_program_version('google-chrome'),
            check_selenium()
        ]
        if not all(prerequisites):
            click.secho('Please install the required dependencies.', fg='red')
            sys.exit(1)


if __name__ == '__main__':
    mtcli()




import warnings
from datetime import datetime, timedelta, timezone

import pandas as pd
import yfinance as yf

from minitrade.datasource import QuoteSource
from minitrade.utils.config import config


class YahooQuoteSource(QuoteSource):
    '''YahooQuoteSource uses `yfinance` to access the financial data available on Yahoo Finance. 

Yahoo Finance offers an excellent range of market data on stocks, bonds, currencies and cryptocurrencies.

Usage:
- Returned data are adjusted for splits and dividends by default. 
- Pass `use_adjusted=False` to `get_source()` to get unadjusted data.

Accepted symbol format:
- U.S. stocks: AAPL, MSFT, BRK-B, ...
- U.S. ETF: SPY, QQQ, ...
- U.S. index: ^GSPC, ^DJI, ^IXIC, ^RUT, ^VIX, ...
- U.S. futures: ES=F, YM=F, NQ=F, RTY=F, CL=F, GC=F, SI=F, HG=F, ...
- U.S. options: TSLA260116C00210000, TSLA260116P00210000, ...
- U.S. mutual fund: VWUSX, ...
- U.S. bond: ^TNX, ^TYX, ^FVX, ^IRX, ...
- U.S. currency: JPY=X, EURUSD=X, GBPUSD=X, ...
- U.S. commodity: CL=F, GC=F, SI=F, HG=F, ...
- Worldwide stocks: 0700.HK, 000001.SS, BMW.DE, VOD.L, ...
- Worldwide ETF: 2800.HK, 510050.SS, ...
- Worldwide index: ^HSI, ^N225, ^FTSE, ...
- Crypto: BTC-USD, ETH-USD, ...

Daily bar:
- The daily OHLCV data includes historical data up to T-1 and the current spot price at T if it is available. Please note that the spot price might not be available for all tickers, could be delayed, and may not be entirely accurate. Therefore, it is advisable to verify the data quality specifically for the tickers you are interested in before use.

Minute bar:
- Experimental

Spot price:
- Experimental
    '''

    def __init__(self, proxy: str = None, use_adjusted=True):
        ''' Yahoo data source

        Parameters
        ----------
        proxy: str
            Http proxy URI to override currently setting if not None
        '''
        self.proxy = proxy or config.sources.yahoo.proxy
        self.use_adjusted = use_adjusted

    def _format_ticker(self, ticker):
        return ticker

    def _ticker_timezone(self, ticker):
        return yf.Ticker(self._format_ticker(ticker)).fast_info['timezone']

    def _ticker_calendar(self, ticker):
        mapping = {
            'WCB': 'NYSE',   # Chicago Board of Options Exchange (CBOE)
            'NMS': 'NYSE',   # National Market System
            'NGM': 'NYSE',   # Nasdaq global market
            'NYQ': 'NYSE',   # NYSE
            'PCX': 'NYSE',   # Pacific Stock Exchange, NYSE Arca
            'CXI': 'NYSE',   # ^VIX
            'PNK': 'NYSE',   # Pink Sheets OTC Market
            'SSH': 'SSE',    # Shanghai stock exchange
            'SHZ': 'SSE',    # Shenzhen stock exchange
            'HKG': 'HKEX',   # Hong Kong stock exchange
            'BTS': 'NYSE',   # BATS Global Markets
        }
        exch = yf.Ticker(self._format_ticker(ticker)).fast_info['exchange']
        if exch in mapping:
            return mapping[exch]
        else:
            warnings.warn(f'Unknown exchange {exch} for {ticker}')
            return 'NYSE'

    def _daily_bar(self, ticker, start, end):
        # Push 1 day out to include "end" in final data
        end_1 = end and (datetime.strptime(end, '%Y-%m-%d') + timedelta(days=1)).strftime('%Y-%m-%d')
        # Yahoo finance uses '-' instead of '.' in ticker symbol
        tk = yf.Ticker(self._format_ticker(ticker))
        df: pd.DataFrame = tk.history(start=start, end=end_1, interval='1d',
                                      auto_adjust=self.use_adjusted, proxy=self.proxy, timeout=10)
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        # Today's data from history api are not reliable. Replace them with spot prices.
        today = self.today(ticker)
        if (not df.loc[today:].empty or not end or end >= today.strftime('%Y-%m-%d')) and self.is_trading_now(ticker):
            tz = df.index.tzinfo
            df = df[df.index < today]
            df.index = df.index.tz_convert(timezone.utc)
            row = dict(zip(df.columns, [
                tk.fast_info['open'],
                tk.fast_info['dayHigh'],
                tk.fast_info['dayLow'],
                tk.fast_info['lastPrice'],
                tk.fast_info['lastVolume']]))
            df = pd.concat([df, pd.DataFrame(row, index=[today.astimezone(timezone.utc)])])
            df.index = df.index.tz_convert(tz)
        # drop timezone info and keep date only
        df.index.rename('Date', inplace=True)
        df.index = df.index.tz_localize(None).normalize()
        return df

    def _minute_bar(self, ticker: str, start: str, end: str, interval: int):
        # Supported intervals: 1, 2, 5, 15, 30, 60
        if interval not in [1, 2, 5, 15, 30, 60]:
            raise ValueError(f'Interval {interval} is not supported')
        # resolution vs range per request, 1m - 7d, 2m - 60d, 5m - 60d, 15m - 60d, 30m - 60d, 60m - 730d
        period = {1: 7, 2: 60, 5: 60, 15: 60, 30: 60, 60: 730}
        # Push 1 day out to include "end" in final data
        end_1 = end and (datetime.strptime(end, '%Y-%m-%d') + timedelta(days=1)).strftime('%Y-%m-%d')
        tk = yf.Ticker(self._format_ticker(ticker))
        if start:
            df: pd.DataFrame = tk.history(start=start, end=end_1, interval=f'{interval}m',
                                          auto_adjust=True, proxy=self.proxy, timeout=10)
        else:
            df: pd.DataFrame = tk.history(period=f'{period[interval]}d', interval=f'{interval}m',
                                          auto_adjust=True, proxy=self.proxy, timeout=10)
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        df.index = df.index.tz_convert(timezone.utc)
        return df

    def _spot(self, tickers):
        try:
            data = {ticker: yf.Ticker(self._format_ticker(ticker)).fast_info['lastPrice']
                    if self.is_trading_now(ticker) else None for ticker in tickers}
            df = pd.Series(data, name=datetime.now(timezone.utc)).astype(float)
            return df
        except Exception as e:
            raise AttributeError(f'Data error') from e



import os
import tempfile

import pandas as pd
import quantstats as qs
import streamlit as st

from minitrade.broker import Broker, BrokerAccount

st.set_page_config(page_title='Broker', layout='wide')


def show_create_broker_account_form():
    brokers = Broker.AVAILABLE_BROKERS
    broker = st.selectbox('Broker', brokers.keys(), format_func=lambda k: brokers[k])
    mode = st.radio('Account type', ['Paper', 'Live'])
    username = st.text_input('Username')
    password = st.text_input('Password', type='password')
    alias = st.text_input('Alias', value=f'{broker}-{mode}-{username}')
    save = st.button('Test and save')
    if save:
        return BrokerAccount(
            alias=alias,
            broker=broker,
            mode=mode,
            username=username,
            password=password,
        )


def test_and_save_broker_account(account: BrokerAccount):
    if account:
        account.save()
        broker = Broker.get_broker(account)
        st.info(
            f'Logging in account {account.username}. Please pay attention to 2FA notification if enabled.')
        try:
            broker.connect()
            st.success('Login succeeded')
        except ConnectionError:
            st.error('Login failed')


def show_broker_account_selector():
    acc_lst = BrokerAccount.list()
    acc_idx = st.sidebar.radio('Broker account', list(range(len(acc_lst))),
                               format_func=lambda i: acc_lst[i].alias)
    account = acc_lst[int(acc_idx)] if acc_idx is not None else None
    return account


def confirm_delete_broker_account(account: BrokerAccount) -> None:
    def confirm_delete():
        if st.session_state.delete_confirm_textinput == account.alias:
            account.delete()
    st.text_input(f'Type "{account.alias}" and press Enter to delete',
                  on_change=confirm_delete, key='delete_confirm_textinput')


def show_broker_account_header_and_controls(account: BrokerAccount):
    c1, c2, c3, c4 = st.columns([6, 1, 1, 1])
    c1.subheader(account.alias)
    c2.button('Refresh')
    if c3.button('Login'):
        broker = Broker.get_broker(account)
        try:
            broker.connect()
            st.success('Login succeeded')
        except ConnectionError:
            st.error('Login failed')
    if c4.button('Delete', type='primary'):
        confirm_delete_broker_account(account)


def show_ib_account_overview(tab, info, portfolio):
    with tab:
        account = st.selectbox(
            'Account', info, format_func=lambda x: f'{x["displayName"]} ({x["accountId"]})', key='ib_account_overview')
        if 'ledger' in account:
            ledger = account['ledger']['BASE']
            c1, c2, c3, c4, c5, c6 = st.columns(6)
            c1.metric('Net Liquidation', f'{ledger.get("netliquidationvalue", 0):,.0f}')
            c2.metric('Cash Balance', f'{ledger.get("cashbalance", 0):,.0f}')
            c3.metric('Interest', f'{ledger.get("interest", 0):,.0f}')
            c4.metric('Dividends', f'{ledger.get("dividends", 0):,.0f}')
            c5.metric('Unrealized PnL', f'{ledger.get("unrealizedpnl", 0):,.0f}')
            c6.metric('Realized PnL', f'{ledger.get("realizedpnl", 0):,.0f}')
            c1.metric('Stock Value', f'{ledger.get("stockmarketvalue", 0):,.0f}')
            c2.metric('Option Value', f'{ledger.get("stockoptionmarketvalue", 0):,.0f}')
            c3.metric('Future Value', f'{ledger.get("futuremarketvalue", 0):,.0f}')
            c4.metric('Future Option Value', f'{ledger.get("futureoptionmarketvalue", 0):,.0f}')
            c5.metric('Warrant Value', f'{ledger.get("warrantsmarketvalue", 0):,.0f}')
            c6.metric('Commodity Value', f'{ledger.get("commoditymarketvalue", 0):,.0f}')
            c1.metric('Bond Value', f'{ledger.get("tbondsmarketvalue", 0):,.0f}')
            c2.metric('Bill Value', f'{ledger.get("tbillsmarketvalue", 0):,.0f}')
            c3.metric('Corporate Bond Value', f'{ledger.get("corporatebondsmarketvalue", 0):,.0f}')
            c4.metric('Fund Value', f'{ledger.get("funds", 0):,.0f}')
            c5.metric('Money Fund Value', f'{ledger.get("moneyfunds", 0):,.0f}')
            c6.metric('Crypto Value', f'{ledger.get("cryptocurrencyvalue", 0):,.0f}')
        if 'performance' in account:
            c1, c2 = st.columns(2)
            c1.caption('NAV')
            nav = account['performance']['nav']
            nav_s = pd.Series(nav['data'][0]['navs'], index=pd.to_datetime(nav['dates'], format='%Y%m%d'))
            c1.line_chart(nav_s, height=300)
            c2.caption('Cumulative PnL')
            cps = account['performance']['cps']
            cps_s = pd.Series(cps['data'][0]['returns'], index=pd.to_datetime(cps['dates'], format='%Y%m%d'))
            c2.line_chart(cps_s, height=300)
        st.caption('Portfolio')
        st.write(portfolio[portfolio['acctId'] == account['id']])


def show_ib_tearsheet(tab, info):
    with tab:
        account = st.selectbox(
            'Account', info, format_func=lambda x: f'{x["displayName"]} ({x["accountId"]})', key='ib_tearsheet')
        if 'performance' in account:
            cps = account['performance']['cps']
            cps_s = pd.Series(cps['data'][0]['returns'], index=pd.to_datetime(cps['dates'], format='%Y%m%d'))
            temp = os.path.join(tempfile.gettempdir(), 'quantstats-tearsheet.html')
            qs.reports.html((cps_s+1).pct_change().dropna(), benchmark='SPY', rf=0.0, display=False,
                            output=temp, title=f'{account["displayName"]} ({account["accountId"]})')
            with open(temp) as f:
                st.components.v1.html(f.read(), height=6000)


def show_broker_account_portfolio_and_trades(account: BrokerAccount):
    tab1, tab2, tab3, tab4, tab5 = st.tabs(['Overview', 'Recent trades', 'Orders', 'Account info', 'Performance'])
    try:
        info, portfolio, trades, orders = get_account_info(account)
        if account.broker == 'IB':
            show_ib_account_overview(tab1, info, portfolio)
            show_ib_tearsheet(tab5, info)
        else:
            tab1.caption('Portfolio')
            tab1.write(portfolio)
            tab5.write('Not available for this account')
        tab2.write(trades)
        tab3.write(orders)
        tab4.write(info)
    except ConnectionError:
        st.info('Login to see account info')


@st.cache_data(ttl='1h')
def get_account_info(account):
    broker = Broker.get_broker(account)
    if broker.is_ready():
        return broker.get_account_info(), broker.get_portfolio(), broker.download_trades(), broker.download_orders()
    else:
        raise ConnectionError('Broker is not ready')


action = st.sidebar.radio('Action', ['Browse existing', 'Add new'])

if action == 'Add new':
    st.subheader('Add broker account')
    account = show_create_broker_account_form()
    if account:
        test_and_save_broker_account(account)

if action == 'Browse existing':
    account = show_broker_account_selector()
    if account:
        show_broker_account_header_and_controls(account)
        show_broker_account_portfolio_and_trades(account)
